/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

/* 
 * File:   ColoredNetStructures.h
 * Author: Klostergaard
 *
 * Created on 17. februar 2018, 17:07
 */

#ifndef COLOREDNETSTRUCTURES_H
#define COLOREDNETSTRUCTURES_H

#include <vector>
#include <set>
#include <algorithm>
#include "Colors.h"
#include "Expressions.h"
#include "Multiset.h"

namespace PetriEngine {
    namespace Colored {
        
        struct Arc {
            uint32_t place;
            uint32_t transition;
            ArcExpression_ptr expr;
            bool input;
            bool skip = false;
        };
        
        struct Transition {
            std::string name;
            GuardExpression_ptr guard;
            std::vector<std::unordered_map<std::string,const Color*>> bindings;
            std::vector<uint32_t> pre;
            std::vector<uint32_t> post;
            //std::vector<size_t> arcs;
            bool skip = false;

            void sort(std::vector<Arc>& arcs) {
                auto less = [arcs](auto i, auto j) {
                    return arcs[i].place < arcs[j].place;
                };
                std::sort(pre.begin(), pre.end(), less);
                std::sort(post.begin(), post.end(), less);
            }

            bool is_sorted(std::vector<Arc>& arcs) {
                auto less = [arcs](auto i, auto j) {
                    return arcs[i].place < arcs[j].place;
                };
                return std::is_sorted(pre.begin(), pre.end(), less)
                       && std::is_sorted(post.begin(), post.end(), less);
            }

            std::vector<uint32_t>::iterator getInArc(std::vector<Arc>& arcs, uint32_t place) {
                auto less = [arcs](auto i, auto j) {
                    return arcs[i].place < j;
                };
                auto ait = std::lower_bound(pre.begin(), pre.end(), place, less);
                if (ait != pre.end() && arcs[*ait].place == place) {
                    return ait;
                }
                return pre.end();
            }

            std::vector<uint32_t>::iterator getOutArc(std::vector<Arc>& arcs, uint32_t place) {
                auto less = [arcs](auto i, auto j) {
                    return arcs[i].place < j;
                };
                auto ait = std::lower_bound(post.begin(), post.end(), place, less);
                if (ait != post.end() && arcs[*ait].place == place) {
                    return ait;
                }
                return post.end();
            }
        };
        
        struct Place {
            std::string name;
            ColorType* type;
            Multiset marking;
            bool skip = false;
            std::vector<uint32_t> producers;
            std::vector<uint32_t> consumers;
        };
    }
}

#endif /* COLOREDNETSTRUCTURES_H */

