//
// Created by andreas on 3/30/18.
//

#ifndef VERIFYPN_COLOREDREDUCER_H
#define VERIFYPN_COLOREDREDUCER_H

#include <glob.h>
#include <iostream>
#include "ColoredNetStructures.h"
#include "../PQL/Contexts.h"

namespace PetriEngine {
    class ColoredPetriNetBuilder;

    namespace Colored {
        class ColoredQueryPlaceAnalysisContext : public PQL::ColoredAnalysisContext {
        private:
            std::vector<uint32_t> _placeInQuery;
            std::vector<uint32_t> _transitionInQuery;

        public:
            ColoredQueryPlaceAnalysisContext(const std::unordered_map<std::string, uint32_t> &places,
                                             const std::unordered_map<std::string, uint32_t> &tnames,
                                             const PetriNet *net,
                                             const std::unordered_map<std::string, std::unordered_map<uint32_t, std::string>> &cplaces,
                                             const std::unordered_map<std::string, std::vector<std::string>> &ctnames,
                                             bool colored);

            bool resolvePlace(const std::string &place, std::unordered_map<uint32_t, std::string> &out) override;
            bool resolveTransition(const std::string &transition, std::vector<std::string> &out) override;

            uint32_t* getQueryPlaceCount() {
                return _placeInQuery.data();
            }

            uint32_t* getQueryTransitionCount() {
                return _transitionInQuery.data();
            }
        };

        class ColoredReducer {
        public:
            ColoredReducer(ColoredPetriNetBuilder*);
            void reduce(ColoredQueryPlaceAnalysisContext& context, int enableReduction, int timeOut, bool remove_loops, bool remove_consumers, bool next_safe);

            size_t removedPlaces() const {
                return _removedPlaces;
            }

            size_t removedTransitions() const {
                return _removedTransitions;
            }

            void printStats(std::ostream& out) const {
                out << "Removed transitions: " << _removedTransitions << "\n"
                    << "Removed places: " << _removedPlaces << "\n"
                    << "Applications of rule I: " << _ruleI << std::endl;
            }

        private:
            size_t _removedPlaces = 0;
            size_t _removedTransitions = 0;
            size_t _ruleI = 0;
            ColoredPetriNetBuilder* parent = nullptr;

            bool reduceByRuleI(uint32_t* placeInQuery, uint32_t* transitionInQuery, bool remove_loops, bool remove_consumers);

            Transition& getTransition(uint32_t transition);
            std::vector<uint32_t>::iterator getOutArc(Transition&, uint32_t place);
            std::vector<uint32_t>::iterator getInArc(uint32_t place, Transition&);
            void eraseTransition(std::vector<uint32_t>& set, uint32_t el);
            void skipPlace(uint32_t);
            void skipTransition(uint32_t);

            bool consistent();

            bool hasTimedOut() const {
                return false; // TODO
            }
        };
    }
}


#endif //VERIFYPN_COLOREDREDUCER_H
